import numpy as np
import pynapple as nap
import pytest
import scipy.signal as signal
from pynajax.jax_process_filtering import jax_sosfiltfilt


def _naive_multiepoch_filtering(sos, tsdframe: nap.TsdFrame):
    """filter epochs in a loop."""
    out = []
    for iset in tsdframe.time_support:
        out.append(signal.sosfiltfilt(sos, tsdframe.restrict(iset).d, axis=0))
    return nap.TsdFrame(t=tsdframe.t, d=np.vstack(out), time_support=tsdframe.time_support)


@pytest.mark.parametrize("wn", [0.1, 0.3])
@pytest.mark.parametrize("order", [2, 4])
@pytest.mark.parametrize("btype", ["lowpass", "highpass"])
@pytest.mark.parametrize("n_sig", [1, 3])
def test_filter_recursion_single_band(wn, order, btype, n_sig):
    sig = np.random.normal(size=(1000, n_sig))
    sos = signal.butter(order, wn, btype=btype, output="sos")
    out_sci = signal.sosfiltfilt(sos, sig, axis=0)
    out_jax = jax_sosfiltfilt(sos, np.arange(sig.shape[0]), sig, [0], [1000])
    assert np.allclose(out_jax, out_sci)


@pytest.mark.parametrize("wn", [[0.1, 0.4], [0.3, 0.5]])
@pytest.mark.parametrize("order", [2, 4])
@pytest.mark.parametrize("btype", ["bandpass", "bandstop"])
@pytest.mark.parametrize("n_sig", [1, 3])
def test_filter_recursion_range(wn, order, btype, n_sig):
    sig = np.random.normal(size=(1000, n_sig))
    sos = signal.butter(order, wn, btype=btype, output="sos")
    out_sci = signal.sosfiltfilt(sos, sig, axis=0)
    out_jax = jax_sosfiltfilt(sos, np.arange(sig.shape[0]), sig, [0], [1000])
    assert np.allclose(out_jax, out_sci)


@pytest.mark.parametrize("wn", [0.1, 0.3])
@pytest.mark.parametrize("order", [2, 4])
@pytest.mark.parametrize("btype", ["lowpass", "highpass"])
@pytest.mark.parametrize("n_sig", [1, 3])
@pytest.mark.parametrize(
    "ep",
    [
        nap.IntervalSet(start=[0], end=[1000]),
        nap.IntervalSet(start=[0, 50], end=[40, 1000]),
        nap.IntervalSet(start=[0, 50, 100], end=[40, 90, 1000])
    ]
)
def test_filter_recursion_single_band_multiepoch_tsdframe(wn, order, btype, n_sig, ep):
    sig = np.random.normal(size=(1000, n_sig))
    sig = nap.TsdFrame(t=np.arange(sig.shape[0]), d=sig, time_support=ep)
    sos = signal.butter(order, wn, btype=btype, output="sos")
    out_sci = _naive_multiepoch_filtering(sos, sig)
    out_jax = jax_sosfiltfilt(sos, sig.t, sig.d, sig.time_support.start, sig.time_support.end)
    assert np.allclose(out_jax, out_sci)


@pytest.mark.parametrize("wn", [0.1, 0.3])
@pytest.mark.parametrize("order", [2, 4])
@pytest.mark.parametrize("btype", ["lowpass", "highpass"])
@pytest.mark.parametrize("n_sig", [1, 3])
@pytest.mark.parametrize(
    "ep",
    [
        nap.IntervalSet(start=[0], end=[1000]),
        nap.IntervalSet(start=[0, 50], end=[40, 1000]),
        nap.IntervalSet(start=[0, 50, 100], end=[40, 90, 1000])
    ]
)
def test_filter_recursion_single_band_multiepoch_tsd(wn, order, btype, n_sig, ep):
    sig = np.random.normal(size=(1000, ))
    sig = nap.Tsd(t=np.arange(sig.shape[0]), d=sig, time_support=ep)
    sos = signal.butter(order, wn, btype=btype, output="sos")
    out_sci = _naive_multiepoch_filtering(sos, sig[:, None])[:, 0]
    out_jax = jax_sosfiltfilt(sos, sig.t, sig.d, sig.time_support.start, sig.time_support.end)
    assert np.allclose(out_jax, out_sci)


@pytest.mark.parametrize("wn", [0.1, 0.3])
@pytest.mark.parametrize("order", [2, 4])
@pytest.mark.parametrize("btype", ["lowpass", "highpass"])
@pytest.mark.parametrize("n_sig", [1, 3])
@pytest.mark.parametrize(
    "ep",
    [
        nap.IntervalSet(start=[0], end=[1000]),
        nap.IntervalSet(start=[0, 50], end=[40, 1000]),
        nap.IntervalSet(start=[0, 50, 100], end=[40, 90, 1000])
    ]
)
def test_filter_recursion_single_band_multiepoch_tsdtensor(wn, order, btype, n_sig, ep):
    sig = np.random.normal(size=(1000, 2, 3))
    sig = nap.TsdTensor(t=np.arange(sig.shape[0]), d=sig, time_support=ep)
    sig_frame = nap.TsdFrame(t=sig.t, d=sig.d.reshape(sig.shape[0], -1), time_support=ep)
    sos = signal.butter(order, wn, btype=btype, output="sos")
    out_sci = _naive_multiepoch_filtering(sos, sig_frame).reshape(sig.shape)
    out_jax = jax_sosfiltfilt(sos, sig.t, sig.d, sig.time_support.start, sig.time_support.end)
    assert np.allclose(out_jax, out_sci)




from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
import scipy.signal as signal

from .utils import (
    _get_shifted_indices,
    _get_slicing,
    _odd_ext_multiepoch,
    _revert_epochs,
)


@partial(jax.jit, static_argnums=(3, ))
def _recursion_loop_sos(signal, sos, zi, nan_function):
    """
    Applies a recursive second-order section (SOS) filter to the input signal.

    Parameters
    ----------
    signal : jnp.ndarray
        The input signal to be filtered, with shape (n_samples,).
    sos : jnp.ndarray
        Array of second-order filter coefficients in the 'sos' format, with shape (n_sections, 6).
    zi : jnp.ndarray
        Initial conditions for the filter, with shape (n_sections, 2, n_epochs).
    nan_function : callable
        A function that specifies how to re-initialize the initial conditions when a NaN is encountered in the signal.
        It should take two arguments: the epoch number and the current filter state, and return a tuple of the updated
        epoch number and the re-initialized filter state.

    Returns
    -------
    jnp.ndarray
        The filtered signal, with the same shape as the input signal.
    """

    def internal_loop(s, x_zi):
        x_cur, zi_slice = x_zi
        x_new = sos[s, 0] * x_cur + zi_slice[s, 0]
        zi_slice = zi_slice.at[s, 0].set(
            sos[s, 1] * x_cur - sos[s, 4] * x_new + zi_slice[s, 1]
        )
        zi_slice = zi_slice.at[s, 1].set(
            sos[s, 2] * x_cur - sos[s, 5] * x_new)
        x_cur = x_new
        return x_cur, zi_slice

    def recursion_step(carry, x):
        epoch_num, zi_slice = carry

        x_cur, zi_slice = jax.lax.fori_loop(
            lower=0, upper=sos.shape[0], body_fun=internal_loop, init_val=(x, zi_slice)
        )

        # Use jax.lax.cond to choose between nan_case and not_nan_case
        epoch_num, zi_slice = jax.lax.cond(
            jnp.isnan(x),  # Condition to check
            nan_function,  # Function to call if x is NaN
            lambda i, x: (i, zi_slice),  # Function to call if x is not NaN
            epoch_num,
            zi,
        )

        return (epoch_num, zi_slice), x_cur

    _, res = jax.lax.scan(recursion_step, (0, zi[..., 0]), signal)

    return res


# vectorize the recursion over signals.
_vmap_recursion_sos = jax.vmap(_recursion_loop_sos, in_axes=(1, None, 2, None), out_axes=1)


def _insert_constant(idx_start, idx_end, data_array, window_size, const=jnp.nan):
    """
    Insert a constant value array between epochs in a time series data array.

    This function interleaves a constant value array of specified size between each epoch in the data array.

    Parameters
    ----------
    idx_start : jnp.ndarray
        Array of start indices for each epoch.
    idx_end : jnp.ndarray
        Array of end indices for each epoch.
    data_array : jnp.ndarray
        The input data array, with shape (n_samples, ...).
    window_size : int
        The size of the constant array to be inserted between epochs.
    const : float, optional
        The constant value to be inserted, by default jnp.nan.

    Returns
    -------
    data_array: jnp.ndarray
        The modified data array with the constant arrays inserted.
    ix_orig: jnp.ndarray
        Indices corresponding to the samples in the original data array.
    ix_shift: jnp.ndarray
        The shifted indices after the constant array has been interleaved.
    idx_start_shift:
        The shifted start indices of the epochs in the modified array.
    idx_end_shift:
        The shifted end indices of the epochs in the modified array.
    """
    # shift by a window every epoch
    idx_start_shift, idx_end_shift = _get_shifted_indices(
        idx_start, idx_end, window_size
    )

    # get the indices for setting elements
    ix_orig = _get_slicing(idx_start, idx_end)
    ix_shift = _get_slicing(idx_start_shift, idx_end_shift)

    tot_size = ix_shift[-1] - ix_shift[0] + 1
    data_array = (
        jnp.full((tot_size, *data_array.shape[1:]), const)
        .at[ix_shift]
        .set(data_array[ix_orig])
    )
    return data_array, ix_orig, ix_shift, idx_start_shift, idx_end_shift


def jax_sosfiltfilt(sos, time_array, data_array, starts, ends):
    """
    Apply forward-backward filtering using a second-order section (SOS) filter.

    This function applies an SOS filter to the data array in both forward and reverse directions,
    which results in zero-phase filtering.

    Parameters
    ----------
    sos : np.ndarray
        Array of second-order filter coefficients in the 'sos' format, with shape (n_sections, 6).
    time_array : np.ndarray
        The time array corresponding to the data, with shape (n_samples,).
    data_array : jnp.ndarray
        The data array to be filtered, with shape (n_samples, ...).
    starts : np.ndarray
        Array of start indices for the epochs in the data array.
    ends : np.ndarray
        Array of end indices for the epochs in the data array.

    Returns
    -------
    : jnp.ndarray
        The zero-phase filtered data array, with the same shape as the input data array.
    """

    original_shape = data_array.shape
    data_array = data_array.reshape(data_array.shape[0], -1)

    # same default padding as scipy.sosfiltfilt ("pad" method and "odd" padtype).
    n_sections = sos.shape[0]
    ntaps = 2 * n_sections + 1
    ntaps -= min((sos[:, 2] == 0).sum(), (sos[:, 5] == 0).sum())
    pad_num = 3 * ntaps

    ext, ix_start_pad, ix_end_pad, ix_data = _odd_ext_multiepoch(pad_num, time_array, data_array, starts, ends)

    # get the start/end index of each epoch after padding
    ix_start_ep = np.hstack((ix_start_pad[0], ix_start_pad[1:-1] + pad_num))
    ix_end_ep = np.hstack((ix_start_ep[1:], ix_end_pad[-1]))

    zi = signal.sosfilt_zi(sos)

    # this braodcast has shape (*zi.shape, data_array.shape[1], len(ix_start_pad))
    z0 = zi[..., jnp.newaxis, jnp.newaxis] * ext.T[jnp.newaxis, jnp.newaxis, ..., ix_start_ep]

    if len(starts) > 1:
        # multi epoch case augmenting with nans.
        aug_data, ix_orig, ix_shift, idx_start_shift, idx_end_shift = _insert_constant(
            ix_start_ep, ix_end_ep, ext, window_size=1, const=np.nan
        )

        # grab the next initial condition, increase the epoch counter
        nan_func = lambda ep_num, x: (ep_num + 1, x[..., ep_num + 1])
    else:
        # single epoch, no augmentation
        nan_func = lambda ep_num, x: (ep_num + 1, x[..., 0])
        aug_data = ext
        idx_start_shift = ix_start_ep
        idx_end_shift = ix_end_ep
        ix_shift = slice(None)


    # call forward recursion
    out = _vmap_recursion_sos(aug_data, sos, z0, nan_func)

    # reverse time axis
    irev = _revert_epochs(idx_start_shift, idx_end_shift)
    out = out.at[ix_shift].set(out[irev])

    # compute new init cond
    z0 = zi[..., jnp.newaxis, jnp.newaxis] * out.T[jnp.newaxis, jnp.newaxis, ..., idx_start_shift]

    # call backward recursion
    out = _vmap_recursion_sos(out, sos, z0, nan_func)

    # re-flip axis
    out = out.at[ix_shift].set(out[irev])

    # remove nans and padding
    out = out[ix_shift][ix_data]

    return out.reshape(original_shape)



import jax
import jax.numpy as jnp
import numpy as np

from pynajax.jax_core_bin_average import bin_average

from .utils import _fill_forward, _get_idxs, _get_shifted_indices, _get_slicing, pad_and_roll

# dot prod shifted counts vs 1D var y vmapped over shift
# [(n_shift, T), (T, )] -> (n_shift, )
_dot_prod = jax.vmap(lambda x, y: jnp.nansum(x * y), in_axes=(0, None), out_axes=0)
# vmap over the neurons
# [(n_shift, T, n_neurons), (T, )] -> (n_shift, n_neurons)
_dot_prod_neu = jax.vmap(_dot_prod, in_axes=(2, None), out_axes=1)
# vmap over the features
# [(n_shift, T, n_neurons), (T, n_features)] -> (n_shift, n_neurons, n_features)
_dot_prod_feature = jax.vmap(_dot_prod_neu, in_axes=(None, 1), out_axes=2)


def event_trigger_average(
    time_target_array,
    count_array,
    time_array,
    data_array,
    starts,
    ends,
    windows,
    binsize,
    batch_size,
):
    """
    Main function to call for event-triggered averages (ETA) across multiple epochs.
    This function assumes count array and data array have been restricted before.

    Parameters
    ----------
    time_target_array : ArrayLike
        Timestamps array of counts
    count_array : ArrayLike
        Count array of events
    time_array : ArrayLike
        Timestamps of data. Can be lower or above sampling rate of time_array
    data_array : ArrayLike
        Data to average at event
    starts : ArrayLike
        Start times of the epochs over which the STA is computed.
    ends : ArrayLike
        End times of the epochs, corresponding to each start time.
    windows : tuple of int
        The number of time steps to include before and after each event for averaging.
    binsize : float
        The bin size. Used by `bin_average`.
    batch_size : int, optional
        Number of elements to process in each batch.

    Returns
    -------
    jnp.ndarray :
        The combined STA calculated across all specified epochs.

    """
    # Need to bring data_target_array to same shape as count_array
    # bin_average
    if count_array.shape[0] < data_array.shape[0]:
        time_array, data_array = bin_average(
            time_array, data_array, starts, ends, binsize
        )
    # fill_forward
    else:
        data_array = _fill_forward(
            time_target_array, time_array, data_array, starts, ends
        )

    idx_start, idx_end = _get_idxs(time_target_array, starts, ends)
    idx_start_shift, idx_end_shift = _get_shifted_indices(
        idx_start, idx_end, np.sum(windows) + 1
    )

    # get the indices for setting elements
    ix_orig = _get_slicing(idx_start, idx_end)
    ix_shift = _get_slicing(idx_start_shift, idx_end_shift)

    # define larger array
    tot_size = ix_shift[-1] - ix_shift[0] + 1

    shape = data_array.shape
    if data_array.ndim == 1:
        data_array = np.expand_dims(data_array, -1)
    else:
        data_array = data_array.reshape(shape[0], -1)

    num_full_batches = np.maximum(1, data_array.shape[1] // batch_size)
    batch_size = np.minimum(data_array.shape[1], batch_size)
    carry = 0

    tot_count = jnp.sum(count_array, 0)

    count_array = (
        jnp.full((tot_size, *count_array.shape[1:]), np.nan)
        .at[ix_shift]
        .set(count_array[ix_orig])
    )
    count_array = pad_and_roll(count_array, windows)

    def scan_fn(carry, x):
        slc = jax.lax.dynamic_slice(
            data_array, (0, carry), (data_array.shape[0], batch_size)
        )
        slc = (
            jnp.full((tot_size, *slc.shape[1:]), np.nan).at[ix_shift].set(slc[ix_orig])
        )
        batch_result = _dot_prod_feature(count_array, slc)
        return carry + batch_size, batch_result

    _, res = jax.lax.scan(scan_fn, carry, None, length=num_full_batches)

    res = jnp.transpose(res, (1, 2, 0, 3))  # move features at the end
    res = res.reshape(*res.shape[:-2], -1)  # flatten features

    extra_elements = data_array.shape[1] % batch_size
    if extra_elements:
        # compute residual slice
        slc = data_array[:, -extra_elements:]
        slc = (
            jnp.full((tot_size, *slc.shape[1:]), np.nan).at[ix_shift].set(slc[ix_orig])
        )
        resid = _dot_prod_feature(count_array, slc)
        # resid = resid.transpose(1, 2, 0).reshape(*res.shape[:-1], -1)
        res = np.concatenate([res, resid], axis=2)

    # reshape back to original
    res = res.reshape((np.sum(windows) + 1, count_array.shape[-1], *shape[1:]))

    res = jnp.apply_along_axis(jnp.divide, 1, res, tot_count)

    return res


def perievent_continuous(data_array, N_w, N_target, slice_idx, w_starts):
    """
    Extract data from data array.

    Parameters
    ----------
    data_array : ArrayLike
        The jax array to extract the data
    N_w : int
        window size
    N_target : int
        Number of reference time points
    slice_idx : ArrayLike
        2-d numpy array of size (N_target, 2). Each row is the start and end of the slice
    w_starts : ArrayLike
        1-d numpy array of size (N_target,). Whether to trim the window on the left when slicing
    """

    if not isinstance(data_array, jnp.ndarray):
        data_array = jnp.asarray(data_array)

    new_data_array = jnp.full((N_w, N_target, *data_array.shape[1:]), jnp.nan)

    w_sizes = slice_idx[:, 1] - slice_idx[:, 0]  # Different sizes

    all_w_sizes = np.unique(w_sizes)
    all_w_start = np.unique(w_starts)

    for w_size in all_w_sizes:
        for w_start in all_w_start:
            col_idx = w_sizes == w_size
            new_idx = np.zeros((w_size, np.sum(col_idx)), dtype=int)
            for i, tmp in enumerate(slice_idx[col_idx]):
                new_idx[:, i] = np.arange(tmp[0], tmp[1])

            new_data_array = new_data_array.at[w_start : w_start + w_size, col_idx].set(
                data_array[new_idx]
            )

    return new_data_array



from importlib.metadata import PackageNotFoundError as _PackageNotFoundError
from importlib.metadata import version as _get_version

try:
    __version__ = _get_version("pynajax")
except _PackageNotFoundError:
    # package is not installed
    pass
# from .jax_core_bin_average import bin_average
# from .jax_core_convolve import convolve



import jax
import jax.numpy as jnp
import numpy as np
from numba import jit


@jit(nopython=True)
def _get_mask_and_count(time_array, starts, ends):
    """
    Calculate mask and count of time points within specified epochs.

    Parameters
    ----------
    time_array : numpy.ndarray
        Array of time points.
    starts : numpy.ndarray
        Start times of the epochs.
    ends : numpy.ndarray
        End times of the epochs.

    Returns
    -------
    ix : numpy.ndarray
        Boolean array where True indicates a time point is within an epoch.
    count : numpy.ndarray
        Array of counts of time points within each epoch.
    """
    # Initialize variables: total number of time points, total epochs,
    # a mask for time points within epochs, and a count of time points per epoch.
    n = len(time_array)
    m = len(starts)
    ix = np.zeros(n, dtype=np.bool_)
    count = np.zeros(m, dtype=np.int64)

    # Epoch index
    k = 0
    # Time point index
    t = 0

    # Advance through epochs until finding one that includes the first time point.
    while ends[k] < time_array[t]:
        k += 1

    # Process all epochs.
    while k < m:
        # Advance the time index until reaching a time point within the current epoch.
        while t < n and time_array[t] < starts[k]:
            t += 1

        # Process all time points within the current epoch.
        while t < n and time_array[t] <= ends[k]:
            ix[t] = True  # Mark time point as within an epoch.
            count[k] += 1  # Increment count for the current epoch.
            t += 1

        # Prepare for next epoch or exit if done.
        k += 1
        if k == m or t == n:
            break

    return ix, count


@jit(nopython=True)
def _get_bin_edges(time_array, starts, ends, binsize):
    """
    Compute bin edges for averaging within specified epochs and identify
    within-epoch time points.

    Parameters
    ----------
    time_array : numpy.ndarray
        Array of time points.
    starts : numpy.ndarray
        Start times of the epochs.
    ends : numpy.ndarray
        End times of the epochs.
    binsize : float
        Size of each bin.

    Returns
    -------
    ix : numpy.ndarray
        Boolean array where True indicates a time point is within an epoch.
    edges : numpy.ndarray
        Array of bin edges.
    in_epoch : numpy.ndarray
        Boolean array indicating if the interval between edges lies within an epoch.
    """
    # Preliminary calculations: mask and count of time points in each epoch.
    ix, count_in_epoch = _get_mask_and_count(time_array, starts, ends)

    # Calculate the total number of bins per epoch and overall edges needed.
    m = starts.shape[0]
    nb_bins = np.zeros(m, dtype=np.int32)
    for k in range(m):
        if (ends[k] - starts[k]) > binsize:
            nb_bins[k] = int(np.ceil((ends[k] + binsize - starts[k]) / binsize))
        else:
            nb_bins[k] = 1

    # Initialize arrays for bin edges and their epoch inclusion status.
    nb = np.sum(nb_bins) + starts.shape[0]
    edges = np.zeros(nb, dtype=np.float64)
    in_epoch = np.ones(nb, dtype=np.bool_)

    k = 0  # Epoch index
    edge_idx = 0  # Edge index
    t = 0  # sample index
    b = 0  # bin index

    while k < m:
        maxb = b + nb_bins[k]
        maxt = t + count_in_epoch[k]
        lbound = starts[k]

        # add all left edges
        while b < maxb:
            xpos = lbound
            if xpos + binsize / 2 > ends[k]:
                break
            else:
                edges[edge_idx] = np.around(xpos, 9)
                lbound += binsize
                b += 1
                edge_idx += 1

        # add final edge of epoch
        edges[edge_idx] = edges[edge_idx - 1] + np.around(binsize, 9)
        in_epoch[edge_idx] = False

        edge_idx += 1

        t = maxt
        k += 1

    return ix, edges[:edge_idx], in_epoch[: edge_idx - 1]


@jax.jit
def jit_average(bins, data, edges):
    # Initialize arrays for sums and counts per bin, and perform bin-wise addition.
    n_bins = len(edges) - 1
    sums = jnp.zeros((n_bins, *data.shape[1:])).at[bins].add(data)
    counts = jnp.zeros(n_bins).at[bins].add(1)

    average = (sums.T / counts).T
    return average


def bin_average(time_array, data_array, starts, ends, binsize):
    """
    Perform bin-averaging of data array based on time array within specified epochs.

    Parameters
    ----------
    time_array : numpy.ndarray
        Array of time points.
    data_array : jax.numpy.ndarray
        Multidimensional data array to be averaged, where the first dimension matches time_array.
    starts : numpy.ndarray
        Start times of the epochs.
    ends : numpy.ndarray
        End times of the epochs.
    binsize : float
        Size of each bin.

    Returns
    -------
    time_array_new : jax.numpy.ndarray
        New time array corresponding to the bin centers.
    data_array_new : jax.numpy.ndarray
        New data array containing the averaged values.
    """
    if not isinstance(data_array, jnp.ndarray):
        data_array = jnp.asarray(data_array)

    # Calculate bin edges and identify time points within epochs for averaging.
    ix, edges, in_epoch = _get_bin_edges(time_array, starts, ends, binsize=binsize)

    # Digitize time points to find corresponding bins, adjusting indices to be 0-based.
    bins = np.digitize(time_array[ix], edges) - 1
    if len(bins):
        average = jit_average(bins, data_array[ix], edges)
        # Create a new time array with bin centers, and filter by in-epoch bins.
        time_array_new = edges[:-1] + binsize / 2
        return time_array_new[in_epoch], average[in_epoch]
    else:
        return np.array([]), jnp.asarray([])



import jax.numpy as jnp

from .utils import _get_idxs, _get_slicing


def threshold(time_array, data_array, starts, ends, thr, method):
    """Threshold function for pynajax

    Parameters
    ----------
    time_array : ArrayLike

    data_array : ArrayLike

    starts : ArrayLike

    ends : ArrayLike

    thr : Number

    method : string


    Returns
    -------
    tuple of ArrayLike
        Description
    """
    if not isinstance(data_array, jnp.ndarray):
        data_array = jnp.asarray(data_array)

    idx_start, idx_end = _get_idxs(time_array, starts, ends)
    idx_slicing = _get_slicing(idx_start, idx_end)

    data_array = data_array[idx_slicing]
    time_array = time_array[idx_slicing]

    if method == "above":
        ix = data_array > thr
    elif method == "below":
        ix = data_array < thr
    elif method == "aboveequal":
        ix = data_array >= thr
    elif method == "belowequal":
        ix = data_array <= thr

    ix2 = jnp.diff(ix * 1)

    new_starts = (
        time_array[1:][ix2 == 1]
        - (time_array[1:][ix2 == 1] - time_array[0:-1][ix2 == 1]) / 2
    )
    new_ends = (
        time_array[0:-1][ix2 == -1]
        + (time_array[1:][ix2 == -1] - time_array[0:-1][ix2 == -1]) / 2
    )

    if ix[0]:  # First element to keep as start
        new_starts = jnp.hstack((jnp.array([time_array[0]]), new_starts))
    if ix[-1]:  # last element to keep as end
        new_ends = jnp.hstack((new_ends, jnp.array([time_array[-1]])))

    return time_array[ix], data_array[ix], new_starts, new_ends



"""
# filtering

This notebook compare the jax implementation of Butterworth filter with [scipy sosfiltfilt](https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.sosfiltfilt.html).

Performances of the `'sinc'` mode can be found in the convolve benchmark as it is the function being called underneath.

⚠️ **Warning:** We do not recommend using GPU for filtering as it is much slower for the moment compared to CPU.


"""
import os
import numpy as np
import pynapple as nap
from time import perf_counter
import matplotlib.pyplot as plt

import warnings
warnings.filterwarnings("ignore")



# %%
# Machine Configuration
import jax
print(jax.devices())

# %%
def get_mean_perf(tsd, mode, n=10):
    tmp = np.zeros(n)
    _ = nap.apply_lowpass_filter(tsd, 0.25 * tsd.rate, mode=mode)
    for i in range(n):
        t1 = perf_counter()
        _ = nap.apply_lowpass_filter(tsd, 0.25 * tsd.rate, mode=mode)
        t2 = perf_counter()
        tmp[i] = t2 - t1
    return [np.mean(tmp), np.std(tmp)]

# %%
# # Increasing number of time points

def benchmark_time_points(mode):
    times = []
    for T in np.arange(1000, 100000, 20000):
        time_array = np.arange(T)/1000
        data_array = np.random.randn(len(time_array))
        startend = np.linspace(0, time_array[-1], T//100).reshape(T//200, 2)
        ep = nap.IntervalSet(start=startend[::2,0], end=startend[::2,1])
        tsd = nap.Tsd(t=time_array, d=data_array)#, time_support=ep)
        times.append([T]+get_mean_perf(tsd, mode))
    return np.array(times)


# %%
# Calling with numba/scipy
nap.nap_config.set_backend("numba")
times_butter_scipy = benchmark_time_points(mode="butter")

# %%
# Calling with jax
nap.nap_config.set_backend("jax")
times_butter_jax = benchmark_time_points(mode="butter")

# %%
# Figure

plt.figure()
for arr, label in zip(
    [times_butter_scipy, times_butter_jax],
    ["Butter (scipy)", "Butter (jax)"],
    ):
    plt.plot(arr[:, 0], arr[:, 1], "o-", label=label)
    plt.fill_between(arr[:, 0], arr[:, 1] - arr[:, 2], arr[:, 1] + arr[:, 2], alpha=0.2)

plt.legend()
plt.xlabel("Number of time points")
plt.ylabel("Time (s)")
plt.title("Butterworth filter low pass")
# plt.show()


# %%
# # Increasing number of dimensions

def benchmark_dimensions(mode):
    times = []
    T = 60000
    for n in np.arange(1, 100, 20):
        time_array = np.arange(T)/1000
        data_array = np.random.randn(len(time_array), n)
        startend = np.linspace(0, time_array[-1], T//100).reshape(T//200, 2)
        ep = nap.IntervalSet(start=startend[::2,0], end=startend[::2,1])
        tsd = nap.TsdFrame(t=time_array, d=data_array, time_support=ep)
        times.append([n]+get_mean_perf(tsd, mode))
    return np.array(times)

# %%
# Calling with numba/scipy
nap.nap_config.set_backend("numba")
dims_butter_scipy = benchmark_dimensions(mode="butter")

# %%
# Calling with jax
nap.nap_config.set_backend("jax")
dims_butter_jax = benchmark_dimensions(mode="butter")

# %%
# Figure


plt.figure()

for arr, label in zip(
    [dims_butter_scipy, dims_butter_jax],
    ["Butter (scipy)", "Butter (jax)"],
    ):
    plt.plot(arr[:, 0], arr[:, 1], "o-", label=label)
    plt.fill_between(arr[:, 0], arr[:, 1] - arr[:, 2], arr[:, 1] + arr[:, 2], alpha=0.2)

plt.legend()
plt.xlabel("Number of dimensions")
plt.ylabel("Time (s)")
plt.title("Butterworth filter low pass")
plt.show()




"""Test configuration script."""
import numpy as np


class MockArray:
    """
    A mock array class designed for testing purposes. It mimics the behavior of array-like objects
    by providing necessary attributes and supporting indexing and iteration, but it is not a direct
    instance of numpy.ndarray.
    """

    def __init__(self, data):
        """
        Initializes the MockArray with data.
        Parameters
        ----------
        data : Union[numpy.ndarray, List]
            A list of data elements that the MockArray will contain.
        """
        self.data = np.asarray(data)
        self.shape = self.data.shape # Simplified shape attribute
        self.dtype = 'float64'  # Simplified dtype; in real scenarios, this should be more dynamic
        self.ndim = self.data.ndim  # Simplified ndim for a 1-dimensional array

    def __getitem__(self, index):
        """
        Supports indexing into the mock array.
        Parameters
        ----------
        index : int or slice
            The index or slice of the data to access.
        Returns
        -------
        The element(s) at the specified index.
        """
        return self.data[index]

    def __iter__(self):
        """
        Supports iteration over the mock array.
        """
        return iter(self.data)

    def __len__(self):
        """
        Returns the length of the mock array.
        """
        return len(self.data)



import numpy as np
import pytest
from pynajax.jax_process_perievent import event_trigger_average
from pynajax.utils import _get_idxs, pad_and_roll


@pytest.mark.parametrize(
    "windows, array, expected_out",
    [
        (
                (1, 0),
                np.array([[1., 2.], [3., 4.], [5., 6.]]),
                np.array([[[1., 2.], [3., 4.], [5., 6.]], [[np.nan, np.nan], [1., 2.], [3., 4.]]])
        ),
        (
                (1, 1),
                np.array([[1., 2.], [3., 4.], [5., 6.]]),
                np.array([[[3., 4.], [5., 6.],[np.nan, np.nan]], [[1., 2.], [3., 4.], [5., 6.]], [[np.nan, np.nan], [1., 2.], [3., 4.]]])
        ),
        (
                (0, 1),
                np.array([[1., 2.], [3., 4.], [5., 6.]]),
                np.array([[[3., 4.], [5., 6.], [np.nan, np.nan]], [[1., 2.], [3., 4.], [5., 6.]]])
        ),
        (
                (2, 0),
                np.array([[1., 2.], [3., 4.], [5., 6.]]),
                np.array([[[1., 2.], [3., 4.], [5., 6.]], [[np.nan, np.nan], [1., 2.], [3., 4.]], [[np.nan, np.nan], [np.nan, np.nan], [1., 2.]]])
        ),
        (
                (2, 2),
                np.array([[1., 2.], [3., 4.], [5., 6.]]),
                np.array([
                    [[5., 6.], [np.nan, np.nan],  [np.nan, np.nan]],
                    [[3., 4.], [5., 6.], [np.nan, np.nan]],
                    [[1., 2.], [3., 4.], [5., 6.]],
                    [[np.nan, np.nan], [1., 2.], [3., 4.]],
                    [[np.nan, np.nan], [np.nan, np.nan], [1., 2.]],
                ])
        ),
        (
                (0, 2),
                np.array([[1., 2.], [3., 4.], [5., 6.]]),
                np.array([
                    [[5., 6.], [np.nan, np.nan],  [np.nan, np.nan]],
                    [[3., 4.], [5., 6.], [np.nan, np.nan]],
                    [[1., 2.], [3., 4.], [5., 6.]]
                ])
        ),
    ]
)
def test_pad_and_roll(windows, array, expected_out):
    out = np.array(pad_and_roll(array, windows))
    np.testing.assert_array_equal(out, expected_out)


def test_get_idxs():
    # Define test input
    time_array = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    starts = np.array([2, 3, 5])
    ends = np.array([2.9, 4.9, 7])

    # Call the function
    idx_start, idx_end = _get_idxs(time_array, starts, ends)

    # Define the expected output
    expected_idx_start = np.array([1, 2, 4])
    expected_idx_end = np.array([2, 4, 7])

    # Assert statements to check the result
    assert np.array_equal(idx_start, expected_idx_start)
    assert np.array_equal(idx_end, expected_idx_end)


@pytest.mark.parametrize(
    "counts, stimuli, expected_out",
    [
        (
                np.array([[0.], [0.], [1.], [0.], [0.]]),
                np.array([0., 0., 2., 0., 0.]),
                np.array([[0.], [2.], [0.]])
        ),
        (
                np.array([[0.], [1.], [1.], [0.], [0.]]),
                np.array([0., 0., 2., 0., 0.]),
                np.array([[0.], [1.], [1.]])
        ),
        (
                np.array([[0.], [0.], [1.], [1.], [0.]]),
                np.array([0., 0., 2., 0., 0.]),
                np.array([[1.], [1.], [0.]])
        ),
        (
                np.array([[0.], [0.], [0.], [0.], [1.]]),
                np.array([0., 0., 0., 2., 0.]),
                np.array([[2.], [0.], [0.]])
        )
    ]
)
def test_event_trigger_average_single_ep(counts, stimuli, expected_out):
    time = np.arange(counts.shape[0])
    windows = (1, 1)
    binsize = 1.
    starts = [0]
    ends = [6]
    out = event_trigger_average(time, counts, time, stimuli, starts, ends, windows, binsize, batch_size=128)
    np.testing.assert_array_equal(out, expected_out)

@pytest.mark.parametrize(
    "counts, stimuli, expected_out",
    [
        (
                np.array([[0.], [0.], [1.], [1.], [0], [0.]]),
                np.array([0., 0., 2., 2., 0., 0.]),
                np.array([[0.], [2.], [0.]])  # this would be [[1], [2], [1]] in a single epoch setting
        )
    ]
)
def test_event_trigger_average_multi_ep(counts, stimuli, expected_out):
    time = np.arange(counts.shape[0])
    windows = (1, 1)
    binsize = 1.
    starts = [0, 3]
    ends = [2, 7]
    out = event_trigger_average(time, counts, time, stimuli, starts, ends, windows, binsize, batch_size=128)
    np.testing.assert_array_equal(out, expected_out)


@pytest.mark.parametrize(
    "counts, stimuli, expected_out",
    [
        (
                np.array([[0.], [0.], [1.], [0.], [0.]]),
                np.zeros((5, 3, 4)),
                np.zeros((3, 1, 3, 4))
        ),
        (
                np.repeat(
                    np.array([[0.], [0.], [1.], [0.], [0.]]),
                    2,1
                    ),
                np.zeros((5, 12)),
                np.zeros((3, 2, 12))
        )
    ]
)
def test_event_trigger_average_batch_size(counts, stimuli, expected_out):
    stimuli[2] = 2.0
    expected_out[1] = 2.0

    time = np.arange(counts.shape[0])
    windows = (1, 1)
    binsize = 1.
    starts = [0]
    ends = [6]
    out = event_trigger_average(
        time, counts, time, stimuli, starts, ends,
        windows, binsize, batch_size=5
        )
    np.testing.assert_array_equal(out, expected_out)



"""Generate the code reference pages and navigation.

See [CCN template repo](https://ccn-template.readthedocs.io/en/latest/notes/03-documentation/) for why.
"""

from pathlib import Path

import mkdocs_gen_files
nav = mkdocs_gen_files.Nav()

for path in sorted(Path("src").rglob("*.py")):
    module_path = path.relative_to("src").with_suffix("")
    doc_path = path.relative_to("src").with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1] == "__main__":
        continue

    nav[parts] = doc_path.as_posix()
    # if the md file name is `module.md`, generate documentation from docstrings
    if full_doc_path.name != 'index.md':
        with mkdocs_gen_files.open(full_doc_path, "w") as fd:
            ident = ".".join(parts)
            fd.write(f"::: {ident}")
    # if the md file name is `index.md`, add the list of modules with hyperlinks
    else:
        this_module_path = Path("src") / path.parent.name
        module_index = ""
        for module_scripts in sorted(this_module_path.rglob("*.py")):
            if "__init__" in module_scripts.name:
                continue
            module_index += f"* [{module_scripts.name.replace('.py', '')}]" \
                            f"({module_scripts.name.replace('.py', '.md')})\n"

        with mkdocs_gen_files.open(full_doc_path, "w") as fd:
            fd.write(module_index)




    mkdocs_gen_files.set_edit_path(full_doc_path, path)

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())